---------------------------------------------------------------------------------------------
-- QHLQuests Class
---------------------------------------------------------------------------------------------
QHLQuests = {}
QHLQuests.__index = QHLQuests;

function QHLQuests:new(id)

	-- Metatable
	local self = {};
	setmetatable(self, QHLQuests);
	
	-- Frame Variables
	self.id = id;

	-- Reference Variables
	self.obj = self;
	self.parent = nil;
	
	-- Quest Variables
	self.quests = {};
	self.quest_log_indexes = {};
	
	-- Global Quest Objectives
	self.ignored = {};
	self.objectives = {};
	self.objective_icons = {};
	self.completed_icons = {};
	self.objectives_need_update = false;
	
	-- Map Quest Objectives
	self.map_objectives = {};
	
	-- Path Variables
	self.path_nodes = {};
	self.path_next = {};
	self.path_current = {};
	self.path_objectives = {};
	self.path_distance = 0;
	self.path_player_world = 0;
	self.path_player_world_x = 0;
	self.path_player_world_y = 0;
	
	-- Ant Trail Variables
	self.ant_dots = {};
	self.ant_phase = 0.0;
	self.ant_cleared = true;	
	
	-- Return
     	return self;
end

local _;

---------------------------------------------------------------------------------------------
-- Table Functions
---------------------------------------------------------------------------------------------
local free_tables = 0;
local recycled_tables = {};
local function CreateTable()
	
	local objective = nil;
	if free_tables > 0 then
		free_tables = free_tables - 1;
		objective = tremove(recycled_tables);
	else
		objective = {};
	end
	return objective;
end

local function RecycleTable(table)

	if table then
		wipe(table);
		setmetatable(table, nil);
		tinsert(recycled_tables, table);
		free_tables = free_tables + 1;
	end
end

---------------------------------------------------------------------------------------------
-- Update Functions
---------------------------------------------------------------------------------------------
function QHLQuests:Update(elapsed, ...)

	if not self.initialized then
	
		-- Initialize Objectives
		self.initialized = true;
		self:UpdateQuests();
		self:SetQuestsForUpdate();
		self:UpdateQuestMaps();
	end
	
	--if not InCombatLockdown() then
	
		-- Update Quests
		self:UpdateQuests();
	
		-- Find Shortest Path
		self:FindShortestPath();

		-- Update Quest Tracker
		self:UpdateQuestTracker();
	--end
end

---------------------------------------------------------------------------------------------
-- Quest Functions
---------------------------------------------------------------------------------------------
function QHLQuests:GetQuest(questID)
	if questID and self.quests then 
		return self.quests[questID];
	end
end

function QHLQuests:GetQuests()
	return self.quests;
end

function QHLQuests:GetObjectives()
	return self.objectives;
end

function QHLQuests:GetMapObjectives()
	return self.map_objectives;
end

function QHLQuests:SetQuestForUpdate(questID)

	if not questID or questID <= 0 then return end
	
	local quest = self.quests[questID];
	if not quest then
		quest = CreateTable();
		quest.questID = questID;
		quest.update_needed = 1;
		self.quests[questID] = quest;
	end
	quest.update_needed = quest.update_needed + 1;
	quest.update_complete = false;
end

function QHLQuests:SetQuestsForUpdate()

	for index, quest in pairs(self.quests) do
		quest.update_needed = quest.update_needed + 1;
		quest.update_complete = false;
	end
end

function QHLQuests:UpdateQuestMaps()

	local maps = CreateTable();
	local map_count = 0;
	
	-- Look For Quests In Other Maps
	for map,floors in pairs(self.map_objectives) do
	
		-- Iterate Through Floors
		for floor,objectives in pairs(floors) do

			-- Iterate Through Objectives
			for index, objective in ipairs(objectives) do

				-- Check For Quest Updates
				local quest = self.quests[objective.questID];
				if quest and quest.update_needed > 0 then
					
					-- Set Map For Update
					if not maps[objective.map] then
						map_count = map_count + 1;
						maps[objective.map] = CreateTable(); 
					end
					
					-- Set Floor For Update
					if not maps[objective.map][quest.floor] then 
						maps[objective.map][quest.floor] = true; 
					end
				end
			end
		end
	end
	
	-- Update Quests
	WorldMapFrame.blockWorldMapUpdate = true;
	for index, quest in pairs(self.quests) do

		if quest.update_needed > 0 then
		
			quest.map, quest.floor = GetQuestWorldMapAreaID(quest.questID);
			
			-- Maps With Floors Start At Floor 1 Instead of 0
			if quest.floor == 0 and QHL.Astrolabe:GetNumFloors(quest.map) > 0 then quest.floor = 1; end
			
			quest.update_needed = quest.update_needed - 1;
			
			-- Set Map For Update
			if not maps[quest.map] then
				map_count = map_count + 1;
				maps[quest.map] = CreateTable(); 
			end
			
			-- Set Floor For Update
			if not maps[quest.map][quest.floor] then 
				maps[quest.map][quest.floor] = true; 
			end
		end	
	end
	WorldMapFrame.blockWorldMapUpdate = nil;
	
	-- Update Maps
	self:UpdateMaps(maps, map_count);
	for map, floors in pairs(maps) do
		RecycleTable(floors);
	end
	RecycleTable(maps);
end

function QHLQuests:UpdateQuest(questID, questLogIndex, questTitle, level, numObjectives, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily)

	local quest = self.quests[questID];
	if not quest then
		quest = CreateTable();
		quest.questID = questID;
		quest.update_needed = 1;
		self.quests[questID] = quest;
	end
	quest.questLogIndex = questLogIndex;
	quest.questTitle = questTitle;
	quest.level = level;
	quest.numObjectives = numObjectives;
	quest.questTag = questTag;
	quest.suggestedGroup = suggestedGroup;
	quest.isHeader = isHeader;
	quest.isCollapsed = isCollapsed;
	quest.isComplete = isComplete;
	quest.isDaily = isDaily;
	if not quest.map then
		WorldMapFrame.blockWorldMapUpdate = true;
		quest.map, quest.floor = GetQuestWorldMapAreaID(questID);
		WorldMapFrame.blockWorldMapUpdate = nil;
	end
end

function QHLQuests:UpdateQuests()

	-- Get Current Quests
	local quest_log_indexes = CreateTable();
	for n=1,GetNumQuestLogEntries() do
	
		-- Get Quest Info
		--local title, level, suggestedGroup, isHeader, isCollapsed, isComplete, frequency, questID, startEvent, displayQuestID, isOnMap, hasLocalPOI, isTask, isStory = GetQuestLogTitle(i)
		local questTitle, level, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(n);
		local _,questTag = GetQuestTagInfo(questID);
		if not isHeader then
		
			local numObjectives = GetNumQuestLeaderBoards(n);
			local requiredMoney = GetQuestLogRequiredMoney(n);
			
			-- Blizzard isComplete Hack
			if ( isComplete and isComplete < 0 ) then
				isComplete = false;
			elseif ( numObjectives == 0 and GetMoney() >= requiredMoney ) then
				isComplete = true;
			end

			-- Update Quest Info
			quest_log_indexes[questID] = n;
			self:UpdateQuest(questID, n, questTitle, level, numObjectives, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily);
		end
	end
	
	-- Remove Old Quests
	for questID, questLogIndex in pairs(self.quest_log_indexes) do
		if not quest_log_indexes[questID] and self.quests[questID] then

			--QHL:PrintAll(1,1,1,"Removing Quest: ", questID);

			-- Remove Quest
			RecycleTable(self.quests[questID]);
			self.quests[questID] = nil;

			-- Remove Quest From Objectives
			self:RemoveQuestFromMapObjectives(questID);
			--self:RemoveQuestFromGlobalObjectives(questID);
			self:RemoveQuestFromPath(questID);
		end
	end
	
	-- Set New Quest Log Indexes
	RecycleTable(self.quest_log_indexes);
	self.quest_log_indexes = quest_log_indexes;
	
	-- Update Global Objectives
	self:UpdateGlobalObjectives();
end

---------------------------------------------------------------------------------------------
-- Objective Functions
---------------------------------------------------------------------------------------------
-- TODO:
function QHLQuests:GetNextObjective()

	return self.path_objectives[1];
end

function QHLQuests:IsObjectiveWithinRange(objective)

	-- Bad Objective
	if not objective or not objective.map then return false; end
	
	-- Check Map
	if QHL:GetSetting("track_by_zone") and objective.map ~= QHL.map then return false; end
	
	-- Check World
	if objective.world == QHL.world then return true; end
	
	-- Must Not Be Within Range
	return false;
end

function QHLQuests:IsObjectiveWithinGroup(objective)

	-- Bad Objective
	if not objective then return false; end
	
	-- Don't Care About Group
	if not QHL:GetSetting("track_by_group") then return true; end
	
	-- Get Group Member Count
	local party_members = GetNumSubgroupMembers();
	local raid_members = GetNumGroupMembers();
	if not UnitInRaid("player") then raid_members = 0; end
	
	-- Check For Dungeon
	if objective.questTag == "Dungeon" and (party_members <= 0 and raid_members <= 0) then 
		return false;
	end
	
	-- Check For Raid
	if objective.questTag == "Raid" and raid_members <= 0 then 
		return false;
	end
	
	-- Check For Group Quests
	if not objective.suggestedGroup or objective.suggestedGroup <= party_members + 1 then
		return true;
	end
	
	-- Must Not Be Within Group
	return false;
end

function QHLQuests:IsObjectiveWithinLevel(objective)

	-- Bad Objective
	if not objective or not objective.level then return false; end
	
	-- Don't Care About Level
	if not QHL:GetSetting("track_by_level") then return true; end
	
	-- Get Offset
	local level_offset = QHL:GetSetting("track_level_offset");
	if not level_offset then return true; end
	
	-- Check Level
	if (objective.level >= UnitLevel("player") - level_offset) 
	and (objective.level <= UnitLevel("player") + level_offset) then 
		return true;
	end
	
	-- Must Not Be Within Level
	return false;
end

function QHLQuests:IsObjectiveTrackable(objective)

	local tracked = false;

	-- Filter Objectives
	if self.quests[objective.questID] 
	and not self:IsObjectiveIgnored(objective) 
	and self:IsObjectiveWithinRange(objective) 
	and self:IsObjectiveWithinLevel(objective) 
	and self:IsObjectiveWithinGroup(objective) then
		tracked = true;
	end
	
	return tracked;
end

function QHLQuests:IgnoreQuest(questID)

	local objective = self:GetObjectiveByQuestID(questID);
	if objective then self:IgnoreObjective(objective); end
end

function QHLQuests:IgnoreObjective(objective)

	-- Ignore Objective If Not Already Ignored
	for index, ignored in ipairs(self.ignored) do
		if ignored.questID == objective.questID then return; end
	end
	
	-- Add Objective To Ignored
	local ignored_objective = CreateTable();
	ignored_objective.questLogIndex = objective.questLogIndex;
	ignored_objective.questID = objective.questID;
	ignored_objective.questTitle = objective.questTitle;
	ignored_objective.level = objective.level;
	ignored_objective.questTag = objective.questTag;
	ignored_objective.suggestedGroup = objective.suggestedGroup;
	ignored_objective.isComplete = objective.isComplete;
	--ignored_objective.isDaily = objective.isDaily;
	ignored_objective.tracked = objective.tracked;
	ignored_objective.map = objective.map;
	ignored_objective.floor = objective.floor;
	ignored_objective.continent = objective.continent;
	--ignored_objective.zone = objective.zone;
	ignored_objective.pos_x = objective.pos_x;
	ignored_objective.pos_y = objective.pos_y;
	ignored_objective.world = objective.world;
	ignored_objective.world_floor = objective.world_floor;
	ignored_objective.world_x = objective.world_x;
	ignored_objective.world_y = objective.world_y;
	ignored_objective.routeID = objective.routeID;
	tinsert(self.ignored, ignored_objective);
	
	-- Remove Objective From Path
	self:RemoveQuestFromPath(ignored_objective.questID);
end

function QHLQuests:UnignoreQuest(questID)

	local objective = self:GetObjectiveByQuestID(questID);
	if objective then self:UnignoreObjective(objective); end
end

function QHLQuests:UnignoreObjective(objective)

	-- Recycle Objectives
	local removed = 0;
	for i=1,#self.ignored do
		local index = i - removed;
		if index <= #self.ignored and objective.questID == self.ignored[index].questID then
			RecycleTable(self.ignored[index]);
			tremove(self.ignored, index);
			removed = removed + 1;
		end
	end
end

function QHLQuests:ResetIgnoredObjectives()

	-- Recycle Objectives
	for index, objective in ipairs(self.ignored) do
		RecycleTable(objective);
	end

	-- Wipe Ignored Objectives
	wipe(self.ignored);
end

function QHLQuests:IsQuestIgnored(questID)

	local objective = self:GetObjectiveByQuestID(questID);
	if objective then return self:IsObjectiveIgnored(objective); 
	else return false; end
end

function QHLQuests:IsObjectiveIgnored(objective)

	for index, ignored in ipairs(self.ignored) do
		if ignored.questID == objective.questID then return true; end
	end
	return false;
end

function QHLQuests:IsQuestOnPath(questID)

	local objective = self:GetObjectiveByQuestID(questID);
	if objective then return self:IsObjectiveOnPath(objective); 
	else return false; end
end

function QHLQuests:IsObjectiveOnPath(objective)

	for index, path_objective in ipairs(self.path_objectives) do
		if path_objective.questID == objective.questID then return true; end
	end
	return false;
end

-- TODO: Multiple?
function QHLQuests:GetObjectiveByRouteID(routeID)

	if self.objectives then
		for index, objective in ipairs(self.objectives) do
			if objective.routeID == routeID then return objective; end
		end
	end
	return nil;
end

-- TODO: Return Multiple
function QHLQuests:GetObjectiveByQuestID(questID)

	if self.objectives then
		for index, objective in ipairs(self.objectives) do
			if objective.questID == questID then return objective; end
		end
	end
	return nil;
end

function QHLQuests:AddGlobalObjective(objective)
	
	-- Create Global Objective
	local global_objective = CreateTable();
	global_objective.questLogIndex = objective.questLogIndex;
	global_objective.questID = objective.questID;
	global_objective.questTitle = objective.questTitle;
	global_objective.level = objective.level;
	global_objective.questTag = objective.questTag;
	global_objective.suggestedGroup = objective.suggestedGroup;
	global_objective.isComplete = objective.isComplete;
	--global_objective.isDaily = objective.isDaily;
	global_objective.tracked = objective.tracked;
	global_objective.map = objective.map;
	global_objective.floor = objective.floor;
	global_objective.continent = objective.continent;
	--global_objective.zone = objective.zone;
	global_objective.pos_x = objective.pos_x;
	global_objective.pos_y = objective.pos_y;
	global_objective.world = objective.world;
	global_objective.world_floor = objective.world_floor;
	global_objective.world_x = objective.world_x;
	global_objective.world_y = objective.world_y;
	global_objective.routeID = objective.routeID;
	
	-- Insert Global Objective
	tinsert(self.objectives, global_objective);
end

function QHLQuests:RemoveQuestFromGlobalObjectives(questID)
	
	-- Copy Objectives
	local global_objectives = CreateTable();
	local removed_objectives = CreateTable();
	for index, objective in pairs(self.objectives) do
		if objective.questID == questID then
			tinsert(removed_objectives, objective);
		else
			tinsert(global_objectives, objective);
		end
	end
	
	-- Recycle Old Objectives
	for index, objective in pairs(removed_objectives) do
		RecycleTable(objective);
	end
	
	-- Recycle Tables
	RecycleTable(removed_objectives);
	RecycleTable(self.objectives);
	self.objectives = global_objectives;
end

function QHLQuests:WipeGlobalObjectives()
	
	-- Recycle Objectives
	for index, objective in ipairs(self.objectives) do
		RecycleTable(objective);
	end
	
	-- Clear Global Objectives Table
	wipe(self.objectives);
end

function QHLQuests:AddPathObjective(objective)
	
	-- Create Global Objective
	local path_objective = CreateTable();
	path_objective.questLogIndex = objective.questLogIndex;
	path_objective.questID = objective.questID;
	path_objective.questTitle = objective.questTitle;
	path_objective.level = objective.level;
	path_objective.questTag = objective.questTag;
	path_objective.suggestedGroup = objective.suggestedGroup;
	path_objective.isComplete = objective.isComplete;
	--path_objective.isDaily = objective.isDaily;
	path_objective.tracked = objective.tracked;
	path_objective.map = objective.map;
	path_objective.floor = objective.floor;
	path_objective.continent = objective.continent;
	--path_objective.zone = objective.zone;
	path_objective.pos_x = objective.pos_x;
	path_objective.pos_y = objective.pos_y;
	path_objective.world = objective.world;
	path_objective.world_floor = objective.world_floor;
	path_objective.world_x = objective.world_x;
	path_objective.world_y = objective.world_y;
	path_objective.routeID = objective.routeID;
	
	-- Insert Path Objective
	tinsert(self.path_objectives, path_objective);
end

function QHLQuests:RemoveQuestFromPath(questID)
	
	-- Copy Objectives
	local path_objectives = CreateTable();
	local removed_objectives = CreateTable();
	for index, objective in pairs(self.path_objectives) do
		if objective.questID == questID then
			tremove(self.path_current, index);
			tinsert(removed_objectives, objective);
		else
			tinsert(path_objectives, objective);
		end
	end

	-- Recycle Old Objectives
	for index, objective in pairs(removed_objectives) do
		RecycleTable(objective);
	end

	-- Recycle Tables
	RecycleTable(removed_objectives);
	RecycleTable(self.path_objectives);
	self.path_objectives = path_objectives;
end

function QHLQuests:WipePathObjectives()
	
	-- Recycle Objectives
	for index, objective in ipairs(self.path_objectives) do
		RecycleTable(objective);
	end
	
	-- Clear Global Objectives Table
	wipe(self.path_objectives);
end

function QHLQuests:AddMapObjective(questLogIndex, questID, questTitle, level, questTag, suggestedGroup, isComplete, isDaily, tracked, continent, pos_x, pos_y, map, floor)

	--if continent then

		-- Create Zone Objective
		local zone_objective = CreateTable();
		zone_objective.questLogIndex = questLogIndex;
		zone_objective.questID = questID;
		zone_objective.questTitle = questTitle;
		zone_objective.level = level;
		zone_objective.questTag = questTag;
		zone_objective.suggestedGroup = suggestedGroup;
		zone_objective.isComplete = isComplete;
		--zone_objective.isDaily = isDaily;
		zone_objective.tracked = tracked;
		zone_objective.map = map;
		zone_objective.floor = floor;
		zone_objective.continent = continent;
		--zone_objective.zone = zone;
		zone_objective.pos_x = pos_x;
		zone_objective.pos_y = pos_y;
		zone_objective.world, zone_objective.world_floor, zone_objective.world_x, zone_objective.world_y = QHL:GetWorldPosition(map, floor, pos_x, pos_y);
		zone_objective.routeID = QHL.TSP:getID(zone_objective.world_x, zone_objective.world_y);

		-- Create Map Objective Table
		if not self.map_objectives[map] then 
			self.map_objectives[map] = {};
		end

		-- Create Floor Objective Table
		if not self.map_objectives[map][floor] then 
			self.map_objectives[map][floor] = {};
		end

		-- Insert Zone Objective
		tinsert(self.map_objectives[map][floor], zone_objective);
	--end
end

function QHLQuests:RemoveQuestFromMapObjectives(questID)

	-- Look For Quests In Other Maps
	for map,floors in pairs(self.map_objectives) do

		-- Iterate Through Floors
		for floor,objectives in pairs(floors) do

			-- Check For Objectives
			if self.map_objectives[map] and self.map_objectives[map][floor] then

				-- Copy Objectives
				local zone_objectives = CreateTable();
				local removed_objectives = CreateTable();
				for index, objective in ipairs(self.map_objectives[map][floor]) do
					if objective.questID == questID then
						tinsert(removed_objectives, objective);
					else
						tinsert(zone_objectives, objective);
					end
				end

				-- Recycle Old Objectives
				for index, objective in pairs(removed_objectives) do
					RecycleTable(objective);
				end

				-- Recycle Tables
				RecycleTable(removed_objectives);
				RecycleTable(self.map_objectives[map][floor]);
				self.map_objectives[map][floor] = zone_objectives;
			end
		end
	end
end

function QHLQuests:WipeMapObjectives(map, floor)

	if self.map_objectives[map] then
	
		if not floor then
		
			-- Wipe Every Floor
			for floor, objectives in pairs(self.map_objectives[map]) do
				for index, objective in ipairs(objectives) do
					RecycleTable(objective);
				end

				-- Clear Floor Objectives Table
				wipe(self.map_objectives[map][floor]);
			end
		
			-- Clear Map Objectives Table
			wipe(self.map_objectives[map]);
			
		elseif self.map_objectives[map][floor] then

			-- Wipe Objectives On Floor
			for index, objective in ipairs(self.map_objectives[map][floor]) do
				RecycleTable(objective);
			end
			
			-- Clear Floor Objectives Table
			wipe(self.map_objectives[map][floor]);
		end
	end
end

--function QHLQuests:WipeZoneObjectives(continent, zone)
function QHLQuests:WipeZoneObjectives(map)

	--local map = QHL.Astrolabe:GetMapID(continent, zone);

	if self.map_objectives[map] then
	
		-- Recycle Objectives
		for floor, objectives in pairs(self.map_objectives[map]) do
			for index, objective in ipairs(objectives) do
				RecycleTable(objective);
			end

			-- Clear Floor Objectives Table
			wipe(self.map_objectives[map][floor]);
		end

		-- Clear Map Objectives Table
		wipe(self.map_objectives[map]);
	end
end

local scripterrors_backup;
local showobjectives_backup;
local battlefieldminimap_update_backup;
local function BackupMapSettings()
	
	-- Disable Script Error Warnings
	if(scripterrors_backup == null) then scripterrors_backup = GetCVar("ScriptErrors"); end
	SetCVar("ScriptErrors", "1");
	
	-- Hook Battlefield Minimap Update To Prevent Updates
	battlefieldminimap_update_backup = BattlefieldMinimap_Update;
	BattlefieldMinimap_Update = function() end;
	
	-- Backup Settings
	showobjectives_backup = GetCVarBool("questPOI");
	if(not showobjectives_backup) then SetCVar("questPOI", 1); end
end

local function RestoreMapSettings()

	-- Unhook Battlefield Minimap Updates
	BattlefieldMinimap_Update = battlefieldminimap_update_backup;

	-- Restore settings
	if(not showobjectives_backup) then SetCVar("questPOI", 0); end
	
	-- Restore Script Error Setting
	SetCVar("ScriptErrors", scripterrors_backup);
end

function QHLQuests:UpdateGlobalObjectives()

	-- Wipe Global Objectives
	self:WipeGlobalObjectives();
	
	-- Look For Quests In Other Maps
	for map,floors in pairs(self.map_objectives) do

		-- Iterate Through Floors
		for floor,objectives in pairs(floors) do

			-- Copy Objectives To Global Objectives
			for index, objective in ipairs(objectives) do
			
				self:AddGlobalObjective(objective);
			end
		end
	end
	
	--QHL:PrintAll(1,1,1,"Completed Updating Objectives");
end

--function QHLQuests:UpdateZoneObjectives(continent, zone)
function QHLQuests:UpdateZoneObjectives(continent, mapid)

	QHL:PrintAll(1,1,1,continent,":",mapid);

	if not QHL.Astrolabe.WorldMapVisible then
	
		-- Wipe Zone Objectives
		--self:WipeZoneObjectives(continent, zone);
		self:WipeZoneObjectives(mapid);
		
		-- Set Map to Zone
		--SetMapZoom(continent, zone);
		SetMapByID(mapid);
		
		-- Update POIs
		local numEntries = QuestMapUpdateAllQuests();
		QuestPOIUpdateIcons();
		
		--local map, floor = GetCurrentMapAreaID(), GetCurrentMapDungeonLevel();
		floor = GetCurrentMapDungeonLevel();
		for i = 1, numEntries do
			
			-- Get Quest Info
			local questID, questLogIndex = QuestPOIGetQuestIDByVisibleIndex(i);
			local questTitle, level, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(questLogIndex);
			local _, questTag = GetQuestTagInfo(questLogIndex);
			local numObjectives = GetNumQuestLeaderBoards(questLogIndex);
			local requiredMoney = GetQuestLogRequiredMoney(questLogIndex);
			
			-- Blizzard isComplete Hack
			if ( isComplete and isComplete < 0 ) then
				isComplete = false;
			elseif ( numObjectives == 0 and GetMoney() >= requiredMoney ) then
				isComplete = true;
			end
			
			-- Get Objective Location
			local _, x, y, objective = QuestPOIGetIconInfo(questID);
			if x and y then
			
				-- Add Objective
				self:AddMapObjective(questLogIndex, questID, questTitle, level, questTag, suggestedGroup, isComplete, isDaily, IsQuestWatched(questLogIndex), continent, x, y, mapid, floor);
			end
		end
		
		return true;
	else
		return false;
	end
end

function QHLQuests:UpdateMapObjectives(map, floor, force)

	if map > 0 and (force or not QHL.Astrolabe.WorldMapVisible or (map == GetCurrentMapAreaID() and floor == GetCurrentMapDungeonLevel())) then
		
		-- Set Map
		if map ~= GetCurrentMapAreaID() or floor ~= GetCurrentMapDungeonLevel() then 
			SetMapByID(map); 
			if floor ~= 0 then SetDungeonMapLevel(floor); end
		end
		local continent = GetCurrentMapContinent();
		
		-- Wipe Zone Objectives
		self:WipeMapObjectives(map, floor);
		
		-- Update POIs
		local numEntries = QuestMapUpdateAllQuests();
		QuestPOIUpdateIcons();
		
		--QHL:PrintAll(1,1,1,"Updating Map: ", map, " - ", numEntries, " (", continent, ":", zone, ")");
		for i = 1, numEntries do
			
			-- Get Quest Info
			local questID, questLogIndex = QuestPOIGetQuestIDByVisibleIndex(i);
			local questTitle, level, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(questLogIndex);
			local _, questTag = GetQuestTagInfo(questLogIndex);
			local numObjectives = GetNumQuestLeaderBoards(questLogIndex);
			local requiredMoney = GetQuestLogRequiredMoney(questLogIndex);
	
			-- Blizzard isComplete Hack
			if ( isComplete and isComplete < 0 ) then
				isComplete = false;
			elseif ( numObjectives == 0 and GetMoney() >= requiredMoney ) then
				isComplete = true;
			end
			
			-- Get Objective Location
			local _, x, y, objective = QuestPOIGetIconInfo(questID);
			if x and y then
			
				-- Add Objective
				--QHL:PrintAll(1,1,1,questTitle,":",map);
				self:AddMapObjective(questLogIndex, questID, questTitle, level, questTag, suggestedGroup, isComplete, isDaily, IsQuestWatched(questLogIndex), continent, x, y, map, floor);
			end
		end
				
		return true;
	else
		return false;
	end
end

function QHLQuests:UpdateMaps(maps, map_count)

	if map_count > 0 then
	
		-- Backup Settings
		BackupMapSettings();
		
		-- Update Zone Objectives
		local map_string = "";
		WorldMapFrame.blockWorldMapUpdate = true;
		for map, floors in pairs(maps) do
		
			map_string = QHL:Concat(map_string, map, ",");
			for floor, update in pairs(floors) do
				self:UpdateMapObjectives(map, floor, true);
			end
		end
		WorldMapFrame.blockWorldMapUpdate = nil;
		
		--QHL:PrintAll(1,1,1,"Updated Maps: ",map_string);

		-- Update Global Objectives
		self:UpdateGlobalObjectives();
		
		-- Restore Settings
		RestoreMapSettings();
	end
end

---------------------------------------------------------------------------------------------
-- Pathing Functions
---------------------------------------------------------------------------------------------
local function InsertPathNode(path, node)

	for index, path_node in ipairs(path) do
		if path_node == node then return; end
	end
	tinsert(path, node);
end

local function PathsHaveSameNodes(path_1, path_2)

	if #path_1 ~= #path_2 then return false; end
	for n=2,#path_1 do
		local found = false;
		for i=2,#path_2 do
			if path_1[n] == path_2[i] then 
				found = true; 
				break; 
			end
		end
		if not found then return false; end
	end
	return true;
end

local function PathsHaveSameQuests(path_1, path_2)

	if #path_1 ~= #path_2 then return false; end
	for n=2,#path_1 do
		local found = false;
		local objective_1 = QHL.quests:GetObjectiveByRouteID(path_1[n]);
		if objective_1 then
			for i=2,#path_2 do
				local objective_2 = QHL.quests:GetObjectiveByRouteID(path_2[i]);
				if objective_2 and objective_1.questID == objective_2.questID then 
					found = true; 
					break; 
				end
			end
		end
		if not found then return false; end
	end
	return true;
end

function QHLQuests:FindShortestPath()

	if QHL.InBrokenInstance then
	
		-- Wipe Current Path
		wipe(self.path_current);

		-- Set Path Variables
		self.path_distance = 0;
		self.path_player_world_x, self.path_player_world_y = 0, 0;

		-- Wipe Path Objectives
		self:WipePathObjectives();

	elseif not QHL.TSP:IsTSPRunning() and not QHL.InBrokenMap then
			
		-- Wipe Route Nodes
		wipe(self.path_next);
		wipe(self.path_nodes);
		
		-- Insert Objectives
		for index, objective in pairs(self.objectives) do

			-- Filter Objectives
			if self:IsObjectiveTrackable(objective) then
				InsertPathNode(self.path_nodes, objective.routeID);
			end
		end
		
		-- Check For Path Nodes
		if #self.path_nodes > 0 then
		
			-- Insert Current Player Position
			self.path_player_world = QHL.world;
			tinsert(self.path_nodes, 1, QHL.TSP:getID(QHL.world_x, QHL.world_y));
		else
			-- Wipe Pathing
			wipe(self.path_nodes);
			wipe(self.path_current);
			self:WipePathObjectives();
			self.path_distance = 0;
			self.path_player_world = QHL.world;
			self.path_player_world_x = QHL.world_x;
			self.path_player_world_y = QHL.world_y;
			
			-- Nothing To Path
			return;
		end

		-- Start Background TSP Processing
		local zone_width, zone_height = QHL:GetMapWidthAndHeight(QHL.world, QHL.world_floor);
		local running, errormsg = QHL.TSP:SolveTSPBackground(self.path_nodes, nil, nil, zone_width, zone_height, nil, self.path_next);
		if (running == 1) then

			-- Set Finish Instructions
			QHL.TSP:SetFinishFunction(function(output, meta, length, iter, timetaken)
			
				-- Compute Current Path Distance (may have changed since last update)
				if QHL.world == self.path_player_world and PathsHaveSameNodes(output, self.path_current) then
				
					if self.path_current[1] then
					
						-- Update Current Path Player Position
						self.path_current[1] = QHL.TSP:getID(QHL.world_x, QHL.world_y);
						
						-- Get Current Path Distance
						local zone_width, zone_height = QHL:GetMapWidthAndHeight(QHL.world, QHL.world_floor);
						self.path_distance = QHL.TSP:PathLength(self.path_current, zone_width, zone_height);
						
						-- Check If New Path Is Shorter (if not don't change paths)
						if self.path_distance <= length + QHL:GetSetting("track_change_offset") then return; end; --QHL:PrintAll(1,1,1,"No Change: ", self.path_distance," <= ", length); return; end
					end
				end
				
				--QHL:PrintAll(1,1,1,"Route Change: ", self.path_distance," -> ", length);

				-- Wipe Current Path
				wipe(self.path_current);
				
				-- Copy New Path Nodes
				for _,node in ipairs(output) do tinsert(self.path_current, node); end
				
				-- Set Path Variables
				self.path_distance = length;
				self.path_player_world_x, self.path_player_world_y = QHL.TSP:getXY(output[1]);
				
				-- Wipe Old Path Objectives
				self:WipePathObjectives();
				
				-- Set New Path Objectives
				for n=2,#self.path_current do
					local objective = self:GetObjectiveByRouteID(self.path_current[n]);
					if objective then self:AddPathObjective(objective); end
				end				
			end);
		end
	end
end

---------------------------------------------------------------------------------------------
-- Quest Tracker Functions
---------------------------------------------------------------------------------------------
function QHLQuests:UpdateQuestTracker()

	-- FIXME
	--[[
	if QHL:GetSetting("track_quests") and not QHL.InBrokenInstance then
		
		-- Set Manual Sort Order
		WatchFrame_SetSorting(nil, WATCHFRAME_SORT_MANUAL);

		-- Remove Quests Out of Range
		for n=GetNumQuestWatches(),1,-1 do

			local index = GetQuestIndexForWatch(n);
			if self.path_objectives[n] then
				local quest = self:GetQuest(self.path_objectives[n].questID);
				if not quest or quest.questLogIndex ~= index then 
					RemoveQuestWatch(index);
				end
			else
				RemoveQuestWatch(index);
			end
		end

		-- Add Quests in Order
		for index, objective in ipairs(self.path_objectives) do
			local quest = self:GetQuest(objective.questID);
			if quest then AddQuestWatch(quest.questLogIndex); end
		end
		if self.path_objectives[1] then SetSuperTrackedQuestID(self.path_objectives[1].questID); end
	end

	-- Force Minimap Tracker To Always Show Next Quest
	if not self.minimap_hooked then

		-- Store Original Version
		self.minimap_hooked = true;
		self.SetSuperTrackedQuestID = SetSuperTrackedQuestID;

		-- Set New Function
		SetSuperTrackedQuestID = function(questID)

			if QHL:GetSetting("track_quests") and QHL.quests.path_objectives[1] and questID == QHL.quests.path_objectives[1].questID then

				-- Update Minimap Tracker
				return QHL.quests.SetSuperTrackedQuestID(QHL.quests.path_objectives[1].questID);
			end
		end
	end
	--]]
end

---------------------------------------------------------------------------------------------
-- Ant Trail Functions
---------------------------------------------------------------------------------------------
local function ClampLine(x1, y1, x2, y2)
  if x1 and y1 and x2 and y2 then
    local x_div, y_div = (x2-x1), (y2-y1)
    local x_0 = y1-x1/x_div*y_div
    local x_1 = y1+(1-x1)/x_div*y_div
    local y_0 = x1-y1/y_div*x_div
    local y_1 = x1+(1-y1)/y_div*x_div
    
    if y1 < 0 then
      x1 = y_0
      y1 = 0
    end
    
    if y2 < 0 then
      x2 = y_0
      y2 = 0
    end
    
    if y1 > 1 then
      x1 = y_1
      y1 = 1
    end
    
    if y2 > 1 then
      x2 = y_1
      y2 = 1
    end
    
    if x1 < 0 then
      y1 = x_0
      x1 = 0
    end
    
    if x2 < 0 then
      y2 = x_0
      x2 = 0
    end
    
    if x1 > 1 then
      y1 = x_1
      x1 = 1
    end
    
    if x2 > 1 then
      y2 = x_1
      x2 = 1
    end
    
    if x1 >= 0 and x2 >= 0 and y1 >= 0 and y2 >= 0 and x1 <= 1 and x2 <= 1 and y1 <= 1 and y2 <= 1 then
      return x1, y1, x2, y2
    end
  end
end

function QHLQuests:ClearAntTrail()

	if not self.ant_cleared then
		for index, dot in ipairs(self.ant_dots) do
			dot:ClearAllPoints();
			dot:Hide();
		end
	end
end

local function CreateDotTexture(parent, id)
	
	local icon = parent:CreateTexture();

	local w, h = 1/8, 1/8;
	local x, y = ((id-1)%8)*w, math.floor((id-1)/8)*h;

	icon:SetTexture("Interface\\AddOns\\QuestHelperLite\\Art\\Icons.tga");
	icon:ClearAllPoints();
	icon:SetDrawLayer("ARTWORK");
	icon:SetBlendMode("BLEND");
	icon:SetTexCoord(x, x+w, y, y+h);
	icon:SetWidth(5);
	icon:SetHeight(5);
	icon:SetVertexColor(0, 0, 0, 0.35);
	icon:Show();
	
	return icon;
end

function QHLQuests:UpdateAntTrail(elapsed)

	if QHL:GetSetting("trail_show") and QHL.Astrolabe.WorldMapVisible and not QHL.InBrokenMap then
	
		local out = 0;
		
		-- Clear Ant Trail
		self:ClearAntTrail();
		self.ant_cleared = false;

		-- Update Phase
		self.ant_phase = self.ant_phase + elapsed + 0.005;
		while self.ant_phase > 1 do self.ant_phase = self.ant_phase - 1; end
		local remainder = self.ant_phase;

		-- Get Map Info
		local w, h = QHL.map_overlay:GetWidth(), -QHL.map_overlay:GetHeight();
		local c, z, m, f = GetCurrentMapContinent(), GetCurrentMapZone(), GetCurrentMapAreaID(), GetCurrentMapDungeonLevel();
		if c and c == 0 then m = 0; end

		-- Get Player Position
		local last_x, last_y = QHL.Astrolabe:TranslateWorldMapPosition(QHL.map, QHL.floor, QHL.pos_x, QHL.pos_y, m, f);

		-- Draw Trails To Each Objective
		for index, objective in ipairs(self.path_objectives) do
		
			-- Get Objective Position
			local new_x, new_y = QHL.Astrolabe:TranslateWorldMapPosition(objective.map, objective.floor, objective.pos_x, objective.pos_y, m, f);
			local x1, y1, x2, y2 = ClampLine(last_x, last_y, new_x, new_y);
			last_x, last_y = new_x, new_y;
			
			if x1 then
			
				local len = math.sqrt((x1-x2)*(x1-x2)*16/9+(y1-y2)*(y1-y2));
				if len > 0.0001 then
				
					local interval = .025/len;
					local p = remainder*interval;
					while p < 1 do
					
						-- Get Dot Texture
						out = out + 1;
						local dot = self.ant_dots[out];
						if not dot then
						
							-- Create New Dot
							dot = CreateDotTexture(QHL.map_overlay, 13);
							dot:SetDrawLayer("BACKGROUND");
							self.ant_dots[out] = dot;
						end
						
						-- Set Dot Position
						dot:Show();
						dot:ClearAllPoints();
						dot:SetPoint("CENTER", QHL.map_overlay, "TOPLEFT", x1*w*(1-p)+x2*w*p, y1*h*(1-p)+y2*h*p);
						p = p + interval;
					end
					remainder = (p-1)/interval;
				end
			end
		end
	else
		-- Clear Ant Trail
		self:ClearAntTrail();
		self.ant_cleared = true;
	end
end

---------------------------------------------------------------------------------------------
-- POI Icon Functions
---------------------------------------------------------------------------------------------
local function POI_GetQuestFrame(questID)
	for i = 1, MAX_NUM_QUESTS do
		local questFrame = _G["WorldMapQuestFrame"..i];
		if ( questFrame and questFrame.questID == questID ) then
			return questFrame;
		end
	end
end

local function POI_FindButtonByQuestId(questID)

	-- Iterate through POI button types
	for _, button_types in pairs(WorldMapPOIFrame.poiTable) do

		-- Iterate through POI buttons
		for _, button in pairs(button_types) do
			if(button.questID == questID) then return button; end
		end
	end
end

local function PopulatePOIMenu(menu, quest)

	local hide = QHL:CreateMenuItem(menu, QHL:GetSetting("arrow") and QHL_LOCALE_MENU_ARROW_HIDE or QHL_LOCALE_MENU_ARROW_SHOW);
	hide:SetFunction(function () QHL:ToggleArrow() end);

	if quest then 
		if QHL.quests:IsQuestIgnored(quest.questID) then
			QHL:CreateMenuItem(menu, QHL_LOCALE_MENU_OBJECTIVE_UNIGNORE):SetFunction(function () QHL.quests:UnignoreQuest(quest.questID); end)
		else
			QHL:CreateMenuItem(menu, QHL_LOCALE_MENU_OBJECTIVE_IGNORE):SetFunction(function () QHL.quests:IgnoreQuest(quest.questID); end)
		end
	end
end

local function POI_OnClick(self, button, ...)

	local quest = QHL.quests:GetQuest(self.questID);
	if quest and button == "RightButton" then
		local menu = QHL:CreateMenu();
		QHL:CreateMenuTitle(menu, quest.questTitle);
		PopulatePOIMenu(menu, quest);
		menu:ShowAtCursor();
	else
		if GetCVarBool("questPOI") and self.quest then
			WorldMapQuestPOI_OnClick(self);
		end
	end
end

function POI_OnEnter(self)
	--WorldMapPOIFrame.allowBlobTooltip = false;
	local quest = QHL.quests:GetQuest(self.questID);
	if quest then
		WorldMapTooltip:Hide();
		WorldMapQuestPOI_SetTooltip(self, quest.questLogIndex);
	end
end

function POI_OnLeave(self)
	WorldMapTooltip:Hide();
	--WorldMapPOIFrame.allowBlobTooltip = true;
end

local function WorldMapPOI_OnClick(self, button, ...)

	local quest = QHL.quests:GetQuest(self.questID);
	if quest and button == "RightButton" then
		local menu = QHL:CreateMenu();
		QHL:CreateMenuTitle(menu, quest.questTitle);
		PopulatePOIMenu(menu, quest);
		menu:ShowAtCursor();
	end
end

local QUEST_POI_ICONS_PER_ROW = 8;
local QUEST_POI_ICON_SIZE = 0.125;
local function DisplayPOI_Icon(icons, parent, name, style, buttonIndex, offsetIndex, questID)

	local buttonName = name.."_"..buttonIndex;
	local poiButton = icons[buttonIndex];
	local index = buttonIndex - 1 + offsetIndex;
	
	if ( not poiButton ) then
	
		if ( style == "numeric") then 
		
			poiButton = CreateFrame("Button", buttonName, _G[parent:GetName()], "QuestPOINumericTemplate");
			
			local yOffset = 0.5 + floor(index / QUEST_POI_ICONS_PER_ROW) * QUEST_POI_ICON_SIZE;
			local xOffset = mod(index, QUEST_POI_ICONS_PER_ROW) * QUEST_POI_ICON_SIZE;
			poiButton.Number:SetTexCoord(xOffset, xOffset + QUEST_POI_ICON_SIZE, yOffset, yOffset + QUEST_POI_ICON_SIZE);
			
		elseif ( style == "completed") then
		
			poiButton = CreateFrame("Button", buttonName, _G[parent:GetName()], "QuestPOICompletedTemplate");
			
			poiButton.FullHighlightTexture:Show();
			poiButton.IconHighlightTexture:Hide();
			poiButton.Icon:SetSize(24, 24);
			poiButton.NormalTexture:SetAlpha(1);
			poiButton.PushedTexture:SetAlpha(1);
		end
		
		poiButton.name = buttonName;
		icons[buttonIndex] = poiButton;
		
		poiButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
		poiButton:SetScript("OnEnter", POI_OnEnter);
		poiButton:SetScript("OnLeave", POI_OnLeave);
		poiButton:SetScript("OnClick", POI_OnClick);	
	end
	
	poiButton.index = index;
	poiButton.questID = questID;
	poiButton.style = style;
	poiButton.used = true;
	
	if ( poiButton.isSelected ) then
		QuestPOI_DeselectButton(poiButton);
	end
	
	poiButton:Show();
	return poiButton;
end

function QHLQuests:UpdateObjectiveIcons()

	-- Show POI Icons
	local completed_count = 0;
	local objective_count = 0;
	for index, objective in ipairs(self.path_objectives) do

		local icon;
		if objective.isComplete then
			completed_count = completed_count + 1;
			icon = DisplayPOI_Icon(self.completed_icons, QHL.map_overlay, "QuestHelperLite_POI", "completed", completed_count, 0, objective.questID);
		else
			objective_count = objective_count + 1;
			icon = DisplayPOI_Icon(self.objective_icons, QHL.map_overlay, "QuestHelperLite_POI", "numeric", objective_count, completed_count, objective.questID);
		end

		QHL.Astrolabe:PlaceIconOnWorldMap(QHL.map_overlay, icon, objective.map, objective.floor, objective.pos_x, objective.pos_y);
		--icon:SetFrameLevel(100 + #self.path_objectives - index);
	end

	-- Hide Unused Icons
	for n = completed_count + 1, #self.completed_icons do
		if self.completed_icons[n] then self.completed_icons[n]:Hide(); end
	end
	for n = objective_count + 1, #self.objective_icons do
		if self.objective_icons[n] then self.objective_icons[n]:Hide(); end
	end
	
	-- Hide Default Objective Icons
	for index, objective in ipairs(self.objectives) do
		local poiButton = POI_FindButtonByQuestId(objective.questID);
		if poiButton then
			if not self:IsObjectiveOnPath(objective) then
				poiButton:RegisterForClicks("LeftButtonUp", "RightButtonUp");
				poiButton:SetScript("OnClick", WorldMapPOI_OnClick);
				poiButton:Show(); 
			else 
				poiButton:Hide();
			end
		end
	end
	--]]
	
	--[[
	-- Iterate through POI button types
	for _, button_types in pairs(WorldMapPOIFrame.poiTable) do

		-- Iterate through POI buttons
		for _, button in pairs(button_types) do
			button:RegisterForClicks("LeftButtonUp", "RightButtonUp");
			button:SetScript("OnClick", WorldMapPOI_OnClick);
		end
	end--]]
end
